#! /usr/bin/env python3.6
# -*- coding: utf8 -*-

#XtremA: Ein Kommandozeilenprogramm zur Bestimmung von lokalen Extremstellen in einem csv-Datensatz.
#Copyright (C) 2021  B. Sc. Felix Krickl

#This program is free software; you can redistribute it and/or
#modify it under the terms of the GNU General Public License
#as published by the Free Software Foundation; either version 2
#of the License, or (at your option) any later version.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Programm	:	XtremA
# Version   :   1.0.1
# lang      :   de
# Autor     :	Felix Krickl

import argparse
import csv
import os
import sys

# strings   ------------------------------------------------------------------------------------------------------------
mtstr = ""
#   XtremA
xtrema_error = "Xtrema: Error:   "
xtrema = "XtremA"
#   help messages
type_h_for_help = "\nGib 'xtrema -h' für Hilfe ein."
help_dlm = """Der Delimiter ist das Trennsymbol, mit dem die einzelnen Werte in der Eingabedatei voneinander
                getrennt sind. Mögliche Werte sind: ['comma', 'space', 'tab']. Der Delimiter aus der Eingabedatei wird
                auch bei der Ausgabedatei wieder verwendet. Default: 'tab'."""
help_inp = """Dateinamen der Eingabedatei mit Dateiendung angeben. Wie die Dateiendung lautet ist egal.
            Wichtig ist die Formatierung innerhalb des Eingabedokuments. Die Endung ist dabei beliebig.
            txt-Dateien zum Beispiel werden auch problemlos eingelesen.
            Nutzungsbeispiel: '-i input.csv'. Die Angabe des input-Arguments und des output-Arguments ist
            obligatorisch."""
help_out = """Dateinamen der Ausgabedatei mit Dateiendung angeben. Wie input Argument."""
help_typ = """Art der Extremstellen, die erkannt werden sollen. Es kann ausgewählt werden aus ['min', 'max', 'inf'].
            'min' bezieht sich auf Minima, 'max' auf Maxima und 'inf' auf Wendepunkte (Inflections).
            Es können auch mehrere Optionen in beliebiger Reihenfolge ausgewählt werden. Die Angabe dieses
            Arguments mit wenigstens einer Option ist obligatorisch. Es existiert kein Default."""
help_usn = """Das using-Argument bezieht sich auf die einzulesenden Spalten aus der Eingabedatei. Beginnend links mit
            eins wird erst die Spalte, welche den Definitionswert enthält angegeben, dann ein Doppelpunkt geschrieben
            und dann die Spalte, welche den Funktionswert enthält angegeben. Nutzungsbeispiel: '-u 3:4'.
            Default: '1:2'."""
#   error messages
col_len_not_equal = xtrema_error + "Die ausgewählten Spalten in der Eingabedatei sind unterschiedlich lang."
corrupt_data = xtrema_error + """In der Eingabedatei sind fehlerhafte Daten.
              Dezimalzahlen müssen mit Dezimalpunkt angegeben werden. Dezimalzahlen mit Dezimalkomma werden nicht
              erkannt.""" + type_h_for_help
corrupt_usn_arg = xtrema_error + "Im using-Argument liegt ein Fehler vor." + type_h_for_help
inputfile_not_found = xtrema_error + """Die angegebene Eingabedatei konnte im Arbeitsverzeichnis nicht gefunden werden.
                    Prüfe auf Rechtschreibfehler und ob du im richtigen Verzeichnis bist.""" + type_h_for_help
rspns_out_of_range = "Deine Antwort konnte nicht erkannt werden." + type_h_for_help
outputfile_allready_exists = "Die angegeben Ausgabedatei existiert bereits. Möchtest du sie überscheiben ? [J/n] "
too_many_dots_in_inputfile = xtrema_error + \
                             "Der Dateiname darf keine Punkte enthalten, abgesehen vom Punkt vor der Dateiendung" + \
                             type_h_for_help
too_many_dots_in_outputfile = xtrema_error + \
                              "Der Dateiname darf keine Punkte enthalten, abgesehen vom Punkt vor der Dateiendung" + \
                              type_h_for_help
too_many_typ_args = xtrema_error + """Es wurden zu viele Optionen für das xtype Argument angegeben.
                      Es dürfen maximal drei Optionen für dieses Argument angegeben werden.""" + type_h_for_help
usr_not_overweriting = "XtremA wird abgebrochen. Gib deinen Befehl mit einer anderen Ausgabedatei erneut ein."
#   neutral messages
builded_difs = "Differentiale berechnet"
input_read = "Eingabedatei eingelesen"
wrote_output = "Ausgabedatei geschrieben"
#   descriptive
xtrema_descr = """description: XtremA ist ein Kommandozeilenprogramm zur Bestimmung von lokalen Extremstellen
                     in einem csv-Datensatz. Autor: B.Sc. Felix Krickl.
                     In der folgenden Argumentbeschreibung kennzeichnen eckige Klammern [] optionale Argumente
                     und geschweifte Klammern {} positionale Argumente.
                     Eine Kommata-getrennte Sequenz kennzeichnet eine Auswahl.
                     Spitze Klammern <> kennzeichnen eine Stelle, an welcher der Benutzer eine eigene,
                     formatgebundene Eingabe einfüllen muss."""
#   variable strings
inp_file = None
out_file = None

# lists     ------------------------------------------------------------------------------------------------------------
dlm_choices = ['comma', 'space', 'tab']
typ_choices = ['min', 'max', 'inf']
col_x = []
col_y = []
col_dif = []
col_ddif = []
col_min = []
col_max = []
col_inf = []
hdl = None      # hdl: headline

# booleans      --------------------------------------------------------------------------------------------------------
hdl_detected = False
usr_is_overwriting = False

# ints
x = 0
y = 1

# Argument Parser   ----------------------------------------------------------------------------------------------------
parser = argparse.ArgumentParser(description=xtrema_descr, prog="xtrema")
parser.add_argument('-d', '--delimiter', nargs='?', default='tab', type=str, choices=dlm_choices,
                    help=help_dlm, dest='dlm')
parser.add_argument('-i', '--input', nargs=1, type=str, required=True, help=help_inp, metavar='{<input.csv>}',
                    dest='inp')
parser.add_argument('-o', '--output', nargs=1, type=str, required=True, help=help_out, metavar='{<output.csv>}',
                    dest='out')
parser.add_argument('-t', '--xtype', nargs='+', type=str, choices=typ_choices, required=True, help=help_typ,
                    dest='typ')
parser.add_argument('-u', '--using', nargs=1, default='1:2', type=str, help=help_usn, metavar='{<X:Y>}', dest='usn')


class C:    # help-class for parsing the arguments of argparse to into script-variables
    pass


c = C()
args = parser.parse_args(namespace=c)
dlm = c.dlm
wd = os.getcwd()
inp = c.inp
out = c.out
inp_file = inp[0]
out_file = out[0]
inp_file_path = wd + "/" + inp_file
out_file_path = wd + "/" + out_file
typ_args_list = c.typ
typ_args = set(typ_args_list)
usn = c.usn


# functions     --------------------------------------------------------------------------------------------------------


def appropriate_delim():     # returns the delimiter asked in the command line as a string fit for csv.reader()

    if dlm == dlm_choices[0]:
        return ','
    elif dlm == dlm_choices[1]:
        return ' '
    elif dlm == dlm_choices[2]:
        return '\t'


def app_mtstr_to_each():	# appends an empty string to every relevant list

    col_dif.append(mtstr); col_ddif.append(mtstr); col_min.append(mtstr); col_max.append(mtstr); col_inf.append(mtstr)


def build_dif():    # builds the first derivative   # always invoked before build_ddif

    n = 0
    while n <= (len(col_x) - 2):
        try:
            col_dif.append(float(col_y[n+1])-float(col_y[n]))
        except TypeError:
            sys.exit(corrupt_data)
        except ValueError:
            sys.exit(corrupt_data)
        n += 1


def build_ddif():   # builds the second derivative  #maybe invoked after build_dif according to typ-args

    col_ddif.append(mtstr)
    n = 1
    col_ddif.extend([float(col_dif[n+1])-float(col_dif[n]) for n in range(len(col_x) - 2)])


def build_min_max():    # searches for the minima and maxima of the given data

    n = 1
    while n <= (len(col_dif) - 2):
        if float(col_dif[n]) > 0 > float(col_dif[n + 1]):
            col_min.append(mtstr); col_max.append(col_y[n])
        elif float(col_dif[n]) < 0 < float(col_dif[n + 1]):
            col_min.append(col_y[n]); col_max.append(mtstr)
        else:
            col_min.append(mtstr); col_max.append(mtstr)
        n += 1

    col_min.append(mtstr); col_max.append(mtstr)


def build_inf():    # searches for the inflections and stationary points of the given data

    col_inf.append(mtstr)
    n = 2
    for n in range(len(col_ddif)-3):
        if float(col_ddif[n+1]) < 0 < float(col_ddif[n+2]) or float(col_ddif[n+1]) > 0 > float(col_ddif[n+2]):
            col_inf.append(col_y[n])
        else:
            col_inf.append(mtstr)

    col_inf.append(mtstr); col_inf.append(mtstr)


def read_inputfile():   # always invoked

    app_mtstr_to_each()

    with open(inp_file, mode='r') as inputfile:
        reader = csv.reader(inputfile, delimiter=appropriate_delim())

        for current_line in reader:
            if current_line[0][0] != '#':
                col_x.append(current_line[x]); col_y.append(current_line[y])
            elif current_line[0][0] == '#':
                hdl_detected = True
                hdl = current_line

    print(input_read)


def calc_xtrema():     # always invoked

    if len(col_x) == len(col_y):
        build_dif()
        if typ_choices[0] in typ_args or typ_choices[1] in typ_args:
            build_min_max()
        if typ_choices[2] in typ_args:
            build_ddif()
            build_inf()
    else:
        sys.exit(col_len_not_equal)

    print(builded_difs)


def write_outputfile():

    typ_indices = []
    xtrema_matrix = []

    xtrema_matrix.append(col_min[:]); xtrema_matrix.append(col_max[:]); xtrema_matrix.append(col_inf[:])

    for xtype in typ_args_list:
        typ_indices.append(typ_choices.index(xtype))

    with open(out_file, 'w') as outputfile:
        writer = csv.writer(outputfile, delimiter=appropriate_delim())
        if hdl_detected:
            to_write_hdl = [hdl[x], hdl[y]]
            to_write_hdl.extend(typ_args_list)
            writer.writerow(to_write_hdl)
        for n in range(len(col_x)):
            to_write_row = [col_x[n], col_y[n]]
            xtrema_values = [xtrema_matrix[xtype][n] for xtype in typ_indices]
            to_write_row.extend(xtrema_values)
            writer.writerow(to_write_row)

    print(wrote_output)


# Error catching    ----------------------------------------------------------------------------------------------------
if len(str(inp).split('.')) != 2:
    sys.exit(too_many_dots_in_inputfile)
if len(str(out).split('.')) != 2:
    sys.exit(too_many_dots_in_outputfile)
if len(typ_args_list) > 3:
    sys.exit(too_many_typ_args)
if len(str(usn).split(':')) != 2:
    sys.exit(corrupt_usn_arg)
try:
    x = int(str(usn).split(':')[0][-1]) - 1    # usn.X starts with 1 -> subtracting one for indexing purposes
    y = int(str(usn).split(':')[1][0]) - 1	# double indices necessary due to format of usn (argparse)
except NameError:
    sys.exit(corrupt_usn_arg)
if os.path.isfile(inp_file_path) is False:
    sys.exit(inputfile_not_found)
if os.path.isfile(out_file_path) is True:
    for _ in range(3):
        overwrite = input(outputfile_allready_exists)
        if overwrite == 'n' or overwrite == 'N':
            sys.exit(usr_not_overweriting)
        elif overwrite == 'j' or overwrite == 'J':
            usr_is_overwriting = True
            break
    if usr_is_overwriting is False:
        sys.exit(rspns_out_of_range)


# main execution    ----------------------------------------------------------------------------------------------------
read_inputfile()
calc_xtrema()
write_outputfile()
sys.exit('Programm ohne XtremA Errors beendet.')
